enum EdgeLabel{ UndoneEdge, MoveEdge, FixEdge, DivideEdge };

class CEdge{
public:
	int            birthStage;
	class CFace   *face[2];
	class CVertex *vertex[2];
	EdgeLabel      label;
	class CVertex *newVertex;
	class CEdge   *pre;
	class CEdge   *post[2];

	CEdge( int birthStage = 0 ){ this->birthStage = birthStage; face[0] = face[1] = NULL; vertex[0] = vertex[1] = NULL; label = FixEdge; newVertex = NULL; pre = post[0] = post[1] = NULL; }
	CVertex *primaryVertex( class CFace &face ); // face.h
	double distance( ogVector &p0, ogVector &p1 );
	//renew
	CEdge( int birthStage, CEdge &preEdge );
	CFace *connectFace( CFace &face ){ return &face == this->face[0] ? this->face[1] : this->face[0]; }
	void renew( int birthStage, ogPlane &boundary );
	void renewPointer(); // face.h
	//reset
	void resetPointer( int deleteStage ); // face.h
	~CEdge(){ if( newVertex != NULL ) delete newVertex; }
	void reset( int deleteStage );
	void print();
};

double CEdge::distance( ogVector &p0, ogVector &p1 ){
	ogLine e( vertex[0]->position, vertex[1]->position );
	double d0 = e.distance( p0 );
	double d1 = e.distance( p1);
	return d0 > d1 ? d0 : d1;
}

CEdge::CEdge( int birthStage, CEdge &preEdge ){
	this->birthStage = birthStage;
	face[0] = preEdge.face[0];
	face[1] = preEdge.face[1];
	vertex[0] = vertex[1] = NULL;
	newVertex = NULL;
	pre = &preEdge;
	post[0] = post[1] = NULL;
}


void CEdge::renew( int birthStage, ogPlane &boundary ){
	if( label != UndoneEdge ) return;
	if( vertex[0]->label == MoveVertex && vertex[1]->label == FixVertex || vertex[1]->label == MoveVertex && vertex[0]->label == FixVertex ){
		label = DivideEdge;
		post[0] = new CEdge( birthStage, *this );
		post[1] = new CEdge( birthStage, *this );
		ogLine edge( vertex[0]->position, vertex[1]->position );
		ogVector position;
		position.intersection( boundary, edge );
		double m = vertex[0]->position.distance( position );
		double n = vertex[1]->position.distance( position );
		ogVector2 texcoord;
		texcoord.divide( vertex[0]->texcoord, vertex[1]->texcoord, m, n );
		newVertex = new CVertex( position, texcoord, birthStage );
		post[0]->vertex[0] = vertex[0];
		post[0]->vertex[1] = newVertex;
		post[1]->vertex[0] = newVertex;
		post[1]->vertex[1] = vertex[1];
		post[0]->label = vertex[0]->label == MoveVertex ? MoveEdge : FixEdge;
		post[1]->label = vertex[1]->label == MoveVertex ? MoveEdge : FixEdge;
	}else if( vertex[0]->label == MoveVertex || vertex[1]->label == MoveVertex ){
		label = MoveEdge;
	}else{
		label = FixEdge;
	}
}

void CEdge::reset( int deleteStage ){
	if( vertex[0]->post != NULL ){ delete vertex[0]->post; vertex[0]->post = NULL; }
	if( vertex[1]->post != NULL ){ delete vertex[1]->post; vertex[1]->post = NULL; }
	if( post[0] != NULL ){ delete post[0]; post[0] = NULL; }
	if( post[1] != NULL ){ delete post[1]; post[1] = NULL; }
	if( newVertex != NULL ){ delete newVertex; newVertex = NULL; }
}

void CEdge::print(){
	printf( "birthStage %d\n", birthStage );
	switch( label ){
	case MoveEdge:   printf( "MoveEdge\n" ); break;
	case FixEdge:    printf( "FixEdge\n" ); break;
	case DivideEdge: printf( "DivideEdge\n" ); break;
	case UndoneEdge: printf( "UndoneEdge\n" ); break;
	}
	if( face[0] == NULL ) printf( "Face0 None\n" ); else printf( "Face0 Exist\n" );
	if( face[1] == NULL ) printf( "Face1 None\n" ); else printf( "Face1 Exist\n" );
	for( int i=0; i<2; i++ ){
		switch( vertex[i]->label ){
		case MoveVertex: printf( "MoveVertex " ); break;
		case FixVertex:  printf( "FixVertex  " ); break;
		case OnAxis:     printf( "OnAxis     " ); break;
		}
		vertex[i]->position.print();
	}
}
