enum FaceLabel{ UndoneFace, FixFace, MoveFace, DivideFace };
enum LabelResult{ AllVertexMove, AllVertexFix, MoveAndFix };
enum RenewResult{ RenewDone, RenewSuccess, RenewFailure };
enum FaceLocation{ BeforeApart, BeforeConnect, Across, BehindConnect, BehindApart };

class CFace{
public:
	int           birthStage;
	int           groupNumber;
	class CEdge **edge;
	int           vertexSize;
	FaceLabel     label;
	class CEdge  *newEdge;
	class CFace  *pre;
	class CFace  *post[2];

	CVertex *vertex( int i ){ return edge[i]->primaryVertex( *this ); }
	ogVector normal(){ ogVector n; n.normal( this->vertex( 2 )->position, this->vertex( 1 )->position, this->vertex( 0 )->position ); return n; }
	CFace *connectFace( int i ){ return edge[i]->connectFace( *this ); }
	CFace( FILE *fp );
	CVertex *closestVertex( double &distance, ogLine &line );
	CVertex *closestVertex( double &distance, ogLine &line, ogVector &except );
	void labelNeighbors( ogVector &position ){ for( int i=0; i<vertexSize; i++ ) this->vertex( i )->labelNeighbors( position ); }
	CVertex *closestVertex( double &distance, ogVector &position );
	CEdge *closestEdge( double &distance, ogVector &p0, ogVector &p1 );
	LabelResult labelRegion( ogPlane &boundary );
	void labelUndone(){ label = UndoneFace; for( int i=0; i<vertexSize; i++ ) edge[i]->label = UndoneEdge; }
	CEdge *neighborEdge( CVertex &vertex, int select ){ for( int i=0; i<vertexSize; i++ ) if( this->vertex( i ) == &vertex ) return edge[(i-1+vertexSize+select)%vertexSize]; return NULL; }
	int include( CVertex &vertex ){ for( int i=0; i<vertexSize; i++ ) if( this->vertex( i ) == &vertex ) return 1; return 0; }
	//renew
	CFace( int birthStage, CFace &preFace );
	void setBoundary( ogPlane *boundary ); // used in overlap() only
	int  ifIncluded( ogPlane *boundary, int boundarySize ); // used in overlap() only
	int overlap( CFace &face );
	RenewResult  renew( int birthStage, int groupNumber, ogLine &axis, double angle, ogPlane &boundary );
	void renewPointer( int groupNumber ){ this->groupNumber = groupNumber; for( int i=0; i<vertexSize; i++ ) edge[i]->renewPointer(); }
	//reset
	void resetPointer( int deleteStage ){ for( int i=0; i<vertexSize; i++ ) edge[i]->resetPointer( deleteStage ); }
	~CFace(){ if( newEdge != NULL ) delete newEdge; delete edge; }
	void reset( int deleteStage, int groupNumber );
	//draw
	FaceLocation whichSide( ogPlane &plane );
	void draw( int textureNumber );
	void draw( ogColor &color );
	void drawBorder( ogColor &borderColor );
	void draw( ogVector &viewLine, ogColor color[], bool texture[] );
	void print();
};

CFace::CFace( FILE *fp ){
	birthStage = 0;
	groupNumber = 0;

	// load vertex positions
	fscanf( fp, "%d", &vertexSize );
	ogVector  position[64];
	int i;
	for( i=0; i<vertexSize; i++ ){
		fscanf( fp, "%lf%lf", &position[i].x, &position[i].y );
		position[i].z = 0.0;
	}
	// calculate texture coordinates
	ogVector2 texcoord[64];
	double xmin, xmax, ymin, ymax;
	xmin = xmax = position[0].x;
	ymin = ymax = position[0].y;
	for( i=1; i<vertexSize; i++ ){
		if     ( position[i].x < xmin ) xmin = position[i].x;
		else if( position[i].x > xmax ) xmax = position[i].x;
		if     ( position[i].y < ymin ) ymin = position[i].y;
		else if( position[i].y > ymax ) ymax = position[i].y;
	}
	for( i=0; i<vertexSize; i++ ){
		texcoord[i].x = ( position[i].x - xmin )/( xmax - xmin );
		texcoord[i].y = ( position[i].y - ymin )/( ymax - ymin );
	}

	edge = new CEdge *[vertexSize];
	for( i=0; i<vertexSize; i++ ){
		edge[i] = new CEdge();
		edge[i]->face[0] = this;
		edge[i]->vertex[0] = new CVertex( position[i], texcoord[i] );
	}
	for( i=0; i<vertexSize; i++ ){
		edge[i]->vertex[1] = edge[(i+1)%vertexSize]->vertex[0];
	}
	newEdge = NULL;
	pre = post[0] = post[1] = NULL;
}

CVertex *CFace::closestVertex( double &distance, ogLine &line ){
	CVertex *closest;
	distance = DBL_MAX;
	for( int i=0; i<vertexSize; i++ ){
		double d = line.distance( this->vertex( i )->position );
		if( d < distance ){
			closest = this->vertex( i );
			distance = d;
		}
	}
	return closest;
}

CVertex *CFace::closestVertex( double &distance, ogLine &line, ogVector &except ){
	CVertex *closest;
	distance = DBL_MAX;
	for( int i=0; i<vertexSize; i++ ){
		double d = line.distance( this->vertex( i )->position );
		if( d < distance && this->vertex( i )->position.distance( except ) > 1.0 ){
			closest = this->vertex( i );
			distance = d;
		}
	}
	return closest;
}

CVertex *CFace::closestVertex( double &distance, ogVector &position ){
	CVertex *closest;
	distance = DBL_MAX;
	for( int i=0; i<vertexSize; i++ ){
		double d = position.distance( this->vertex( i )->position );
		if( d < distance ){
			closest = this->vertex( i );
			distance = d;
		}
	}
	return closest;
}

CEdge *CFace::closestEdge( double &distance, ogVector &p0, ogVector &p1 ){
	CEdge *closest;
	distance = DBL_MAX;
	for( int i=0; i<vertexSize; i++ ){
		double d = edge[i]->distance( p0, p1 );
		if( d < distance ){
			closest = edge[i];
			distance = d;
		}
	}
	return closest;
}

LabelResult CFace::labelRegion( ogPlane &boundary ){
	int m, b, f;
	m = b = f = 0;
	for( int i=0; i<vertexSize; i++ ){
		switch( this->vertex( i )->labelRegion( boundary ) ){
			case MoveCandidate: m++;break;
			case FixVertex: f++;break;
			case OnAxis: b++;break;
		}
	}
	if( (b+f) == 0 )  return AllVertexMove;
	else if( m == 0 ) return AllVertexFix;
	else              return MoveAndFix;
}

CFace::CFace( int birthStage, CFace &preFace ){
	this->birthStage = birthStage;
	edge = new CEdge *[preFace.vertexSize+1];
	this->vertexSize = 0;
	groupNumber = preFace.groupNumber;
	newEdge = NULL;
	pre = &preFace;
	post[0] = post[1] = NULL;
}


void CFace::setBoundary( ogPlane *boundary ){
	ogVector faceNormal;
	faceNormal.normal( this->vertex( 0 )->position, this->vertex( 1 )->position, this->vertex( 2 )->position );
	for( int i=0; i<vertexSize; i++ ){
		ogVector edge;
		edge.subtract( this->vertex( (i+1)%vertexSize )->position, this->vertex( i )->position );
		boundary[i].normal.exterior( faceNormal, edge );
		boundary[i].normal.normalize();
		boundary[i].passage = this->vertex( (i+1)%vertexSize )->position;
	}
}

int CFace::ifIncluded( ogPlane *boundary, int boundarySize ){
	ogVector gravity( 0.0, 0.0, 0.0 );
	int i;
	for( i=0; i<vertexSize; i++ ) gravity += this->vertex( i )->position;
	gravity /= double( vertexSize );
	for( i=0; i<boundarySize; i++ ) if( boundary[i].signedDistance( gravity ) <= Margin.faceOverlap ) return 0;
	return 1;
}

int CFace::overlap( CFace &face ){
ogPlane *boundary[2];
	this->setBoundary( 	boundary[0] = new ogPlane[vertexSize] );
	face.setBoundary( boundary[1] = new ogPlane[face.vertexSize] );
	// One polygon includes the center of gravity of the other polygon.
	if( this->ifIncluded( boundary[1], face.vertexSize ) ) return 1;
	if( face.ifIncluded( boundary[0], vertexSize ) ) return 1;
	// One of lines of one polygon intersects with that of the other polygon.
	for( int j=0; j<face.vertexSize; j++ ){
		for( int i=0; i<vertexSize; i++ ){
			double distance[4];
			distance[0] = boundary[0][i].signedDistance( face.vertex( j )->position );
			distance[1] = boundary[0][i].signedDistance( face.vertex( (j+1)%face.vertexSize )->position );
			distance[2] = boundary[1][j].signedDistance( this->vertex( i )->position );
			distance[3] = boundary[1][j].signedDistance( this->vertex( (i+1)%vertexSize )->position );
			int intersect = 1;
			if( distance[0]*distance[1] >= 0.0 ) intersect = 0;
			if( fabs(distance[0]) < Margin.faceOverlap ) intersect = 0;
			if( fabs(distance[1]) < Margin.faceOverlap ) intersect = 0;
			if( distance[2]*distance[3] >= 0.0 ) intersect = 0;
			if( fabs(distance[2]) < Margin.faceOverlap ) intersect = 0;
			if( fabs(distance[3]) < Margin.faceOverlap ) intersect = 0;
			if( intersect ) return 1;
		}
	}
	return 0;
}

RenewResult CFace::renew( int birthStage, int groupNumber, ogLine &axis, double angle, ogPlane &boundary ){
	if( label != UndoneFace ) return RenewDone;
	if( groupNumber != this->groupNumber ) return RenewFailure;
	//renew Vertex
	int i;
	for( i=0; i<vertexSize; i++ ){
		if( vertex( i )->label == MoveCandidate ){
			vertex( i )->label = MoveVertex;
			vertex( i )->post = new CVertex( birthStage, *vertex( i ), axis, angle );
		}
	}
	//renew edges
	for( i=0; i<vertexSize; i++ ) edge[i]->renew( birthStage, boundary );
	//create post faces
	int m=0, f=0;
	for( i=0; i<vertexSize; i++ ){
		if( vertex( i )->label == MoveVertex )      m++;
		else if( vertex( i )->label == FixVertex ) f++;
	}
	if( f == 0 ){
		label = MoveFace;
	}else if( m == 0 ){
		label = FixFace;
	}else{
		label = DivideFace;
		post[0] = new CFace( birthStage, *this );
		post[1] = new CFace( birthStage, *this );
		newEdge = new CEdge( birthStage );
		newEdge->label = FixEdge;
		newEdge->face[0] = post[0];
		newEdge->face[1] = post[1];
		for( int i=0; i<vertexSize; i++ ){
			switch( edge[i]->label ){
			case MoveEdge:
				post[0]->edge[post[0]->vertexSize++] = edge[i];
				if( edge[(i+1)%vertexSize]->label == FixEdge ){
					newEdge->vertex[0] = this->vertex( (i+1)%vertexSize );
					post[0]->edge[post[0]->vertexSize++] = newEdge;
				}
				break;
			case FixEdge:
				post[1]->edge[post[1]->vertexSize++] = edge[i];
				if( edge[(i+1)%vertexSize]->label == MoveEdge ){
					newEdge->vertex[1] = this->vertex( (i+1)%vertexSize );
					post[1]->edge[post[1]->vertexSize++] = newEdge;
				}
				break;
			case DivideEdge:
				post[0]->edge[post[0]->vertexSize++] = edge[i]->post[0]->label == MoveEdge ? edge[i]->post[0] : edge[i]->post[1];
				post[1]->edge[post[1]->vertexSize++] = edge[i]->post[0]->label == FixEdge  ? edge[i]->post[0] : edge[i]->post[1];
				if( this->vertex( i )->label == MoveVertex ){
					newEdge->vertex[0] = edge[i]->newVertex;
					post[0]->edge[post[0]->vertexSize++] = newEdge;
				}else{
					newEdge->vertex[1] = edge[i]->newVertex;
					post[1]->edge[post[1]->vertexSize++] = newEdge;
				}
				break;
			}
		}
	}
	for( i=0; i<vertexSize; i++ ){
		if( edge[i]->label == MoveEdge || edge[i]->label == DivideEdge ){
			if( this->connectFace( i ) != NULL ){
				if( this->connectFace( i )->renew( birthStage, groupNumber, axis, angle, boundary ) == RenewFailure ) return RenewFailure;
			}
		}
	}
	return RenewSuccess;
}

void CFace::reset( int deleteStage, int groupNumber ){
	this->groupNumber = groupNumber;
	for( int i=0; i<vertexSize; i++ ) edge[i]->reset( deleteStage );
	if( post[0] != NULL ) { delete post[0]; post[0] = NULL; }
	if( post[1] != NULL ) { delete post[1]; post[1] = NULL; }
	if( newEdge != NULL ) { delete newEdge; newEdge = NULL; }
}

FaceLocation CFace::whichSide( ogPlane &plane ){
	int before=0, behind=0, onIntersection=0;
	for( int i=0; i<vertexSize; i++ ){
		double d = plane.signedDistance( this->vertex( i )->position );
		if( d > Margin.axis ) before = 1;
		else if( d < -Margin.axis ) behind = 1;
		else onIntersection = 1;
	}
	if( before == 1 && behind == 0 ) return onIntersection ? BeforeConnect : BeforeApart;
	else if( before == 0 && behind == 1 ) return onIntersection ? BehindConnect : BehindApart;
	else return Across;
}

void CFace::draw( int textureNumber ){
	ogVFxyzct *v;
	v = new ogVFxyzct[vertexSize];
	for(int i=0;i<vertexSize;i++){
		v[i].set(this->vertex(i)->position);
		v[i].setColor(255, 255, 255);
		v[i].setTexCoord(this->vertex(i)->texcoord);
	}
	og.DrawTexturePolygon(v, vertexSize, textureNumber);
}

void CFace::draw( ogColor &color ){
	ogVFxyzc *v;
	v = new ogVFxyzc[vertexSize];
	for(int i=0;i<vertexSize;i++){
		v[i].set(this->vertex(i)->position);
		v[i].setColor(color.r, color.g, color.b);
	}
	og.DrawPolygon(v,vertexSize);
	delete v;
}

void CFace::drawBorder( ogColor &color ){
	ogVFxyzc *v;
	v = new ogVFxyzc[vertexSize+1];
	for(int i=0;i<=vertexSize;i++){
		v[i].set(this->vertex(i%vertexSize)->position);
		v[i].setColor(color.r, color.g, color.b);
	}
	og.DrawLineStrip(v,vertexSize+1);
	delete v;
}

void CFace::draw( ogVector &frontNormal, ogColor color[], bool texture[] ){
	ogVector n = this->normal();
	if( frontNormal * n > 0 ){
		if( texture[0] ) this->draw( 0 ); else this->draw( color[0] );
	}else{
		if( texture[1] ){
			if(texture[0]) this->draw( 1 ); else this->draw( 0 );
		} else{
			this->draw( color[1] );
		}
	}
	this->drawBorder( color[2] );
//	this->print();
}

void CFace::print(){
	printf( "birthStage %d\n", birthStage );
	printf( "groupNumber %d\n", groupNumber );
	printf( "vertexSize %d\n", vertexSize );
	switch( label ){
		case MoveFace  : printf( "label MoveFace\n" ); break;
		case FixFace   : printf( "label FixFace\n" ); break;
		case DivideFace: printf( "label DivideFace\n" ); break;
		case UndoneFace: printf( "label UndoneFace\n" ); break;
	}
	for( int i=0; i<vertexSize; i++ ){
		switch( this->vertex(i)->label ){
		case MoveVertex: printf( "MoveVertex " ); break;
		case FixVertex:  printf( "FixVertex  " ); break;
		case OnAxis:     printf( "OnAxis     " ); break;
		}
		this->vertex(i)->position.print();
	}
}

// ------------------------- CEdge class member functions -----------------------------------------
CVertex *CEdge::primaryVertex( class CFace &face ){ return &face == this->face[0] || face.pre == this->face[0] || face.post[0] == this->face[0] || face.post[1] == this->face[0] ? vertex[0] : vertex[1]; }

void CEdge::renewPointer(){
	for( int i=0; i<2; i++ ){
		if( vertex[i]->post !=NULL ) vertex[i] = vertex[i]->post;
		if( face[i] != NULL && face[i]->post[0] !=NULL ){
			if( label == MoveEdge ) face[i] = face[i]->post[0];
			else face[i] = face[i]->post[1];
		}
	}
}

void CEdge::resetPointer( int deleteStage ){
	for( int i=0; i<2; i++ ){
		if( vertex[i]->birthStage == deleteStage ) vertex[i] = vertex[i]->pre;
		if( face[i] != NULL && face[i]->birthStage == deleteStage ) face[i] = face[i]->pre;
	}
}


