enum FoldType { FoldUp, FoldDown, TuckIn, BendUp, BendDown, Bend, Fold };

class CFaceGroup{
public:
	int     number;
	int     birthStage;
	CFace **face;
	int     faceSize;
	ogVector normal;

	CFaceGroup( FILE *fp ){ number = 0; birthStage = 0; face = new CFace *[faceSize = 1]; face[0] = new CFace( fp ); normal.unitZ(); }// for the original face
	CVertex *closestVertex( double &distance, ogLine &line );
	CVertex *closestVertex( double &distance, ogLine &line, ogVector &except );
	CVertex *closestVertex( double &distance, ogVector &position );
	CFace **pickedFaceList( CVertex &pickedVertex );
	CEdge *closestEdge( double &distance, ogVector &p0, ogVector &p1 );
	void labelNeighbors( ogVector &position ){ for( int i=0; i<faceSize; i++ ) face[i]->labelNeighbors( position ); }
	LabelResult labelRegion( ogPlane &boundary );
	void labelUndone(){ for( int i=0; i<faceSize; i++ ) face[i]->labelUndone(); }
	//renew
	RenewResult renewFoldUp( int birthStage, ogLine &axis, double angle, ogPlane &boundary );
	RenewResult renewFoldDown( int birthStage, ogLine &axis, double angle, ogPlane &boundary );
	RenewResult renewTuckIn( int birthStage, ogLine &axis, double angle, ogPlane &boundary );
	void renewPointer(){ for( int i=0; i<faceSize; i++ ) face[i]->renewPointer( number ); }
	void constructFoldUp( CFaceGroup &preFaceGroup );
	void constructFoldDown( CFaceGroup &preFaceGroup );
	void constructBend0( CFaceGroup &preFaceGroup );
	void constructBend1( CFaceGroup &preFaceGroup );
	void constructTuckIn( CFaceGroup &preFaceGroup );
	CFaceGroup( int birthStage, CFaceGroup &preFaceGroup, FoldType foldType, int groupNumber = -1 );
	//reset
	~CFaceGroup(){ delete face; }
	void reset( int deleteStage );
	//draw
	int visibleSide( ogVector &eyePosition ){ ogVector n = face[0]->vertex( 0 )->position - eyePosition; return normal * n < 0 ? 0 : 1; }
	FaceLocation whichSide( ogPlane &plane );
	int sort( CFaceGroup &faceGroup, ogVector &eyePosition );
	void draw( ogVector &eyePosition, ogColor color[], bool texture[] );
};

CVertex *CFaceGroup::closestVertex( double &distance, ogLine &line ){
	CVertex *closest;
	distance = DBL_MAX;
	for( int i=0; i<faceSize; i++ ){
		double d;
		CVertex *v = face[i]->closestVertex( d, line );
		if( d < distance ){
			closest = v;
			distance = d;
		}
	}
	return closest;
}

CVertex *CFaceGroup::closestVertex( double &distance, ogLine &line, ogVector &except ){
	CVertex *closest;
	distance = DBL_MAX;
	for( int i=0; i<faceSize; i++ ){
		double d;
		CVertex *v = face[i]->closestVertex( d, line, except );
		if( d < distance ){
			closest = v;
			distance = d;
		}
	}
	return closest;
}

CVertex *CFaceGroup::closestVertex( double &distance, ogVector &position ){
	CVertex *closest;
	distance = DBL_MAX;
	for( int i=0; i<faceSize; i++ ){
		double d;
		CVertex *v = face[i]->closestVertex( d, position );
		if( d < distance ){
			closest = v;
			distance = d;
		}
	}
	return closest;
}

CFace **CFaceGroup::pickedFaceList( CVertex &pickedVertex ){
	int n = 0, i;
	for( i=0; i<faceSize; i++ ) if( face[i]->include( pickedVertex ) ) n++;
	CFace **faceList = new CFace *[n+1];
	n = 0;
	for( i=0; i<faceSize; i++ ) if( face[i]->include( pickedVertex ) ) faceList[n++] = face[i];
	faceList[n++] = NULL;
	return faceList;
}

CEdge *CFaceGroup::closestEdge( double &distance, ogVector &p0, ogVector &p1 ){
	CEdge *closest;
	distance = DBL_MAX;
	for( int i=0; i<faceSize; i++ ){
		double d;
		CEdge *e = face[i]->closestEdge( d, p0, p1 );
		if( d < distance ){
			closest = e;
			distance = d;
		}
	}
	return closest;
}

LabelResult CFaceGroup::labelRegion( ogPlane &boundary ){
	int allVertexMove = 1;
	int allVertexFix  = 1;
	for( int i=0; i<faceSize; i++ ){
		switch( face[i]->labelRegion( boundary ) ){
			case AllVertexMove: allVertexFix = 0;break;
			case AllVertexFix : allVertexMove = 0;break;
			case MoveAndFix   : allVertexMove = allVertexFix = 0;break;
		}
	}
	return allVertexMove ? AllVertexMove : allVertexFix ? AllVertexFix : MoveAndFix;
}

RenewResult CFaceGroup::renewFoldUp( int birthStage, ogLine &axis, double angle, ogPlane &boundary ){
	int renewDone = 1;
	for( int i=0; i<faceSize; i++ ){
		for( int j=i+1; j<faceSize; j++ ){
			//face[i]->print();
			//face[j]->print();
			if( face[i]->label == MoveFace && face[i]->overlap( *face[j] ) || face[i]->label == DivideFace && face[i]->post[0]->overlap( *face[j] ) ){
				switch( face[j]->renew( birthStage, number, axis, angle, boundary ) ){
					case RenewSuccess: renewDone = 0; break;
					case RenewFailure: return RenewFailure;
				}
			}
		}
	}
	return renewDone ? RenewDone : RenewSuccess;
}

RenewResult CFaceGroup::renewFoldDown( int birthStage, ogLine &axis, double angle, ogPlane &boundary ){
	int renewDone = 1;
	for( int i=faceSize-1; i>=0; i-- ){
		for( int j=i-1; j>=0; j-- ){
			if( face[i]->label == MoveFace && face[i]->overlap( *face[j] ) || face[i]->label == DivideFace && face[i]->post[0]->overlap( *face[j] ) ){
				switch( face[j]->renew( birthStage, number, axis, angle, boundary ) ){
					case RenewSuccess: renewDone = 0; break;
					case RenewFailure: return RenewFailure;
				}
			}
		}
	}
	return renewDone ? RenewDone : RenewSuccess;
}

RenewResult CFaceGroup::renewTuckIn( int birthStage, ogLine &axis, double angle, ogPlane &boundary ){
	int renewDone = 1;

	int lowest = 0;
	while( face[lowest]->label != MoveFace && face[lowest]->label != DivideFace && lowest < faceSize-1 ) lowest++;
	int highest = faceSize-1;
	while( face[highest]->label != MoveFace && face[highest]->label != DivideFace && highest > 0 ) highest--;
	//printf("l %d h %d\n",lowest, highest );
	for( int i=lowest+1; i<highest; i++ ){
		if( face[lowest]->label == MoveFace && face[lowest]->overlap( *face[i] ) || face[lowest]->label == DivideFace && face[lowest]->post[0]->overlap( *face[i] ) ){
			switch( face[i]->renew( birthStage, number, axis, angle, boundary ) ){
				case RenewSuccess: renewDone = 0; break;
				case RenewFailure: return RenewFailure;
			}
		}
	}
	return renewDone ? RenewDone : RenewSuccess;
}

void CFaceGroup::constructFoldUp( CFaceGroup &preFaceGroup ){
	normal = preFaceGroup.normal;
	face = new CFace *[preFaceGroup.faceSize*2];
	faceSize = 0;
	int i;
	for( i=0; i<preFaceGroup.faceSize; i++ ){
		if( preFaceGroup.face[i]->label == FixFace || preFaceGroup.face[i]->label == UndoneFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[1];
		}
	}
	for( i=preFaceGroup.faceSize-1; i>=0; i-- ){
		if( preFaceGroup.face[i]->label == MoveFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[0];
		}
	}
}

void CFaceGroup::constructFoldDown( CFaceGroup &preFaceGroup ){
	normal = preFaceGroup.normal;
	face = new CFace *[preFaceGroup.faceSize*2];
	faceSize = 0;
	int i;
	for( i=preFaceGroup.faceSize-1; i>=0; i-- ){
		if( preFaceGroup.face[i]->label == MoveFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[0];
		}
	}
	for( i=0; i<preFaceGroup.faceSize; i++ ){
		if( preFaceGroup.face[i]->label == FixFace || preFaceGroup.face[i]->label == UndoneFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[1];
		}
	}
}

void CFaceGroup::constructBend0( CFaceGroup &preFaceGroup ){
	normal = preFaceGroup.normal;
	face = new CFace *[preFaceGroup.faceSize];
	faceSize = 0;
	for( int i=0; i<preFaceGroup.faceSize; i++ ){
		if( preFaceGroup.face[i]->label == FixFace || preFaceGroup.face[i]->label == UndoneFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[1];
		}
	}
}

void CFaceGroup::constructBend1( CFaceGroup &preFaceGroup ){
	normal = preFaceGroup.normal;
	face = new CFace *[preFaceGroup.faceSize];
	faceSize = 0;
	for( int i=0; i<preFaceGroup.faceSize; i++ ){
		if( preFaceGroup.face[i]->label == MoveFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[0];
		}
	}
}

void CFaceGroup::constructTuckIn( CFaceGroup &preFaceGroup ){
	normal = preFaceGroup.normal;
	face = new CFace *[preFaceGroup.faceSize*2];
	faceSize = 0;
	int lowest = 0;
	while( preFaceGroup.face[lowest]->label != MoveFace && preFaceGroup.face[lowest]->label != DivideFace ) lowest++;
	int highest = preFaceGroup.faceSize-1;
	while( preFaceGroup.face[highest]->label != MoveFace && preFaceGroup.face[highest]->label != DivideFace ) highest--;
	int middle = ( lowest + highest ) / 2;

	int i;
	for( i=0; i<=middle; i++ ){
		if( preFaceGroup.face[i]->label == FixFace || preFaceGroup.face[i]->label == UndoneFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[1];
		}
	}
	for( i=middle; i>=0; i-- ){
		if( preFaceGroup.face[i]->label == MoveFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[0];
		}
	}
	for( i=preFaceGroup.faceSize-1; i>middle; i-- ){
		if( preFaceGroup.face[i]->label == MoveFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[0];
		}
	}
	for( i=middle+1; i<preFaceGroup.faceSize; i++ ){
		if( preFaceGroup.face[i]->label == FixFace || preFaceGroup.face[i]->label == UndoneFace ){
			face[faceSize++] = preFaceGroup.face[i];
		}else if( preFaceGroup.face[i]->label == DivideFace ){
			face[faceSize++] = preFaceGroup.face[i]->post[1];
		}
	}
}

CFaceGroup::CFaceGroup( int birthStage, CFaceGroup &preFaceGroup, FoldType foldType, int groupNumber ){
	this->number = groupNumber == -1 ? preFaceGroup.number : groupNumber;
	this->birthStage = birthStage;
	switch( foldType ){
		case FoldUp: this->constructFoldUp( preFaceGroup ); break;
		case FoldDown: this->constructFoldDown( preFaceGroup ); break;
		case TuckIn: this->constructTuckIn( preFaceGroup ); break;
		case BendUp: groupNumber == -1 ? this->constructBend0( preFaceGroup ) : this->constructBend1( preFaceGroup ); break;
		case BendDown: groupNumber == -1 ? this->constructBend0( preFaceGroup ) : this->constructBend1( preFaceGroup ); break;
	}
	this->renewPointer();
}

void CFaceGroup::reset( int deleteStage ){
	int i;
	for( i=0; i<faceSize; i++ ) face[i]->resetPointer( deleteStage );
	for( i=0; i<faceSize; i++ ) face[i]->reset( deleteStage, number );
}

FaceLocation CFaceGroup::whichSide( ogPlane &plane ){
	int before=0, behind=0, onIntersection=0;
	for( int i=0; i<faceSize; i++ ){
		switch( face[i]->whichSide( plane ) ){
			case BeforeApart: before = 1; break;
			case BeforeConnect: before = onIntersection = 1; break;
			case Across: before = behind = 1; break;
			case BehindConnect: behind = onIntersection = 1; break;
			case BehindApart: behind = 1; break;
		}
	}
	if( before == 1 && behind == 0 ) return onIntersection ? BeforeConnect : BeforeApart;
	else if( before == 0 && behind == 1 ) return onIntersection ? BehindConnect : BehindApart;
	else return Across;
}

int CFaceGroup::sort( CFaceGroup &faceGroup, ogVector &eyePosition ){
	ogPlane plane;
	plane.normal = faceGroup.visibleSide( eyePosition ) == 0 ? faceGroup.normal : -faceGroup.normal;
	plane.passage = faceGroup.face[0]->vertex( 0 )->position;
	FaceLocation side0 = this->whichSide( plane );
	plane.normal = this->visibleSide( eyePosition ) == 0 ? normal : -normal;
	plane.passage = face[0]->vertex( 0 )->position;
	FaceLocation side1 = faceGroup.whichSide( plane );
	//printf( "side0 %d side1 %d\n", side0, side1 );
	if( (side0==BeforeApart || side0==BeforeConnect) && (side1==Across || side1==BehindConnect) ) return 0;
	if( (side1==BehindApart || side1==BehindConnect) && (side0==Across || side0==BeforeConnect) ) return 0;
	return 1;
}

void CFaceGroup::draw( ogVector &eyePosition, ogColor color[], bool texture[] ){
	if( (face[0]->vertex( 0 )->position - eyePosition) * normal < 0.0 ){
		for( int i=0; i<faceSize; i++ ) face[i]->draw( normal, color, texture );
	}else{
		ogVector frontNormal = - normal;
		for( int i=faceSize-1; i>=0; i-- ) face[i]->draw( frontNormal, color, texture );
	}
}



