//Origami Simulator C++ version 5/13/2004

#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "facegroup.h"
#include "stage.h"
#include "origami.h"

COrigami Origami;

class CGUI{
public:
	int        cursorX;
	int        cursorY;
	double     cursorDepth;
	bool       leftMouseOn;
	bool       tuckIn;
	FoldResult foldResult;
	bool       menuChanged;
	bool       drawCoordinateAxis;
	bool       rotation;
	bool       animation;
	double     frameInterval;
	bool       callDisplay;

	CGUI(){
		leftMouseOn = false;
		foldResult = FoldRest;
		drawCoordinateAxis = false;
		rotation = false;
		animation = false;
		frameInterval = 0.0;
		callDisplay = false;
	}
	void setCursorPosition( int x, int y ){ cursorX = x; cursorY = y; }
} GUI;

void initialize(){
	Origami.setup( "origami.txt" );
}


void reshape( int width, int height ){
	og.Viewport( 0, 0, width, height ); 
	og.ZBufferOff();
	og.GouraudShading();
	og.LightOff();
	og.Perspective( 45.0, 1.0, 10000.0);
	og.LookAt( 0, 0, 40,  0, 0, 0,  0, 1, 0 ); //eye, center, up
}

void display( bool drawAdjustMarkMode ){
	og.Clear(0,0,0);
	ogVector eye = og.GetEyePosition();
	Origami.draw( eye );
	if( GUI.drawCoordinateAxis ) Origami.drawCoordinateAxis();
	if( drawAdjustMarkMode ) Origami.drawAdjustMark();

	og.SwapBuffers();
}

void display(){ display( GUI.foldResult == FoldSuccess ); }

void idle(){
	GUI.frameInterval = og.GetFrameInterval();

	if( GUI.leftMouseOn ){
		if( GUI.foldResult == FoldSuccess ) Origami.reset();
		FoldType foldTypeRequest = Fold;
		if( GUI.tuckIn ){
			foldTypeRequest = TuckIn;
		}else if( og.KeyPressed( OGK_M ) ){
			GUI.cursorDepth -= 0.002 * GUI.frameInterval;
			foldTypeRequest = Bend;
		}else if( og.KeyPressed( OGK_N ) ){
			GUI.cursorDepth += 0.002 * GUI.frameInterval;
		}
		ogVector movedPosition;
		og.Window2World( GUI.cursorX, GUI.cursorY, GUI.cursorDepth, movedPosition );
		ogVector eye = og.GetEyePosition();
		GUI.foldResult = Origami.renew( movedPosition, eye, foldTypeRequest, !og.KeyPressed( OGK_F ) );
		og.World2Window( movedPosition, GUI.cursorX, GUI.cursorY, GUI.cursorDepth );
		og.Redisplay();
	}else{
		if( GUI.rotation ){ og.RotateEyePositionLeft( GUI.frameInterval * rad(90.0) ); }
		if( GUI.animation ){
			if( GUI.foldResult == FoldProgress ) Origami.reset();
			GUI.foldResult = Origami.renew( GUI.frameInterval * rad(45.0) );
		}
		if( GUI.callDisplay || og.KeyControlEyePosition( GUI.frameInterval * rad(90.0), GUI.frameInterval * 3.0 ) || GUI.rotation || GUI.animation ){
			og.Redisplay();
			if( GUI.callDisplay ) GUI.callDisplay = false;
		}
		if( GUI.animation && GUI.foldResult == FoldComplete ){ GUI.foldResult = FoldRest; GUI.animation = 0; }
	}
}	

void mouse( int button, int state, int x, int y ){
	if( button == OGK_LBUTTON ){
		if( state == OGK_PRESSED  ){
			if( GUI.rotation || GUI.animation ) return;
			ogVector movedPosition;
			ogLine line = og.CursorLine( x, y );
			Origami.pick( movedPosition, line );
			og.World2Window( movedPosition, GUI.cursorX, GUI.cursorY, GUI.cursorDepth );
			og.WarpPointer( GUI.cursorX, GUI.cursorY );

			GUI.leftMouseOn = true;
			GUI.foldResult = FoldFailure;
			GUI.tuckIn = false;
		}else{
			if( GUI.rotation || GUI.animation ) return;
			GUI.leftMouseOn = false;
			GUI.foldResult = FoldRest;
		}
	}
}

void motion( int x, int y ){ if( GUI.leftMouseOn ) GUI.setCursorPosition( x, y ); }

void menu( int n );

void keyboard( unsigned char key, int x, int y ){
	og.SetKeyState(key,true);
	switch( key ){
		case 't': GUI.tuckIn = !GUI.tuckIn; break;
		case 'q': menu(1); break;
		case 'r': menu(2); break;
		case 'a': menu(3); break;
		case 'z': menu(7); break;
		case 'i': menu(8); break;
	}
	GUI.callDisplay = true;
}

void keyboardUp( unsigned char key, int x, int y ){
	og.SetKeyState(key,false);
}

void special( int key, int x, int y ){
	og.SetKeyState(key,true);
}

void specialUp( int key, int x, int y ){
	og.SetKeyState(key,false);
}

void menu( int n ){
	switch( n ){
		case 1: exit(0);
		case 2:
			if( !GUI.leftMouseOn ) GUI.rotation = !GUI.rotation;
			if( GUI.rotation ) GUI.frameInterval = ogGetFrameInterval();
			break;
		case 3: GUI.drawCoordinateAxis = !GUI.drawCoordinateAxis; break;
		case 4: Origami.save( "default.ori" ); break;
		case 5: Origami.load( "default.ori" ); break;
		case 6:
			if( !GUI.leftMouseOn ) GUI.animation = !GUI.animation;
			if( GUI.animation ){
				Origami.save( "fold.tmp" );
				Origami.reset( "fold.tmp" );
				GUI.frameInterval = ogGetFrameInterval();
			}else{
				Origami.load( "fold.tmp" );
			}
			break;
		case 7: Origami.reset(); break;
		case 8: Origami.reset( 0 ); break;
	}
	GUI.callDisplay = true;
}

