class COrigami{
public:
	// state sequences of origami
	int     maxStage;
	CStage *stage[256];
	// face appearance
	ogColor color[3];   // 0:front 1:back 2:border
	bool    texture[2]; //true:texture false:non-texture
	// animation parameters
	FILE  *animationFp;
	CStage animationStage;
	double animationPosition;

	COrigami(){ }
	void setup( char *filename );
	void pick( ogVector &movedPosition, ogLine &cursorLine ){ stage[maxStage]->pick( movedPosition, cursorLine ); }
	FoldResult renew( ogVector &movedPosition, ogVector &eyePosition, FoldType foldTypeRequest, int adjustMode );
	void reset();
	void reset( int maxStage ){ while( this->maxStage > maxStage ) this->reset(); }
	//draw
	void draw( ogVector &eyePosition ){ stage[maxStage]->draw( eyePosition, color, texture ); }
	void drawAdjustMark(){ stage[maxStage-1]->drawMark(); }
	void drawCoordinateAxis();
	int  defaultSetup( char *filename );
	// file io
	void save( char *filename );
	void load( char *filename );
	// animation
	void reset( char *filename );
	FoldResult renew( double deltaPosition );
};

void COrigami::setup( char *filename ){
	FILE *fp = fopen( filename, "r" );// Open paper definition file
	if( fp  == NULL ){
		if( this->defaultSetup( filename ) ){
			printf( "Can't create %s\n", filename );
			exit( 1 );
		}
		fp = fopen( filename, "r" );
	}

	maxStage = 0;
	
	char   str[64], str2[64];
	while(fscanf(fp,"%s",str) != EOF){
		if(!strcmp(str, "VertexSize")){
			stage[0] = new CStage( fp );//construct an original state
		}else if(!strcmp(str, "FrontColor")){
			fscanf( fp, "%u%u%u", &color[0].r, &color[0].g, &color[0].b );
			texture[0] = false;
		}else if(!strcmp(str, "BackColor")){
			fscanf( fp, "%u%u%u", &color[1].r, &color[1].g, &color[1].b );
			texture[1] = false;
		}else if(!strcmp(str, "BorderColor")){
			fscanf( fp, "%u%u%u", &color[2].r, &color[2].g, &color[2].b );
		}else if(!strcmp(str, "FrontTexture")){
			fscanf(fp, "%s", str2);
			og.LoadTexture( str2 );
			texture[0] = true;
		}else if(!strcmp(str, "BackTexture")){
			fscanf(fp, "%s", str2);
			og.LoadTexture( str2 );
			texture[1] = true;
		}
	}
	fclose( fp );
}

FoldResult COrigami::renew( ogVector &movedPosition, ogVector &eyePosition, FoldType foldTypeRequest, int adjustMode ){
	if( stage[maxStage]->renew( movedPosition, eyePosition, foldTypeRequest, adjustMode ) == FoldFailure ){
		return FoldFailure;
	}else{// folding success
		stage[maxStage+1] = new CStage( *stage[maxStage] );
		maxStage++;
		return FoldSuccess;
	}
}

void COrigami::reset(){
	if( maxStage == 0 ) return;
	maxStage--;
	delete stage[maxStage+1];
	stage[maxStage]->reset();
}

void COrigami::drawCoordinateAxis(){
	og.DrawLine(-20.0,0.0,0.0, 255,0,0, 0.0,0.0,0.0, 255,255,255);
	og.DrawLine( 20.0,0.0,0.0, 255,0,0, 0.0,0.0,0.0, 255,255,255);
	og.DrawLine(0.0,-20.0,0.0, 0,255,0, 0.0,0.0,0.0, 255,255,255);
	og.DrawLine(0.0, 20.0,0.0, 0,255,0, 0.0,0.0,0.0, 255,255,255);
	og.DrawLine(0.0,0.0,-20.0, 0,0,255, 0.0,0.0,0.0, 255,255,255);
	og.DrawLine(0.0,0.0, 20.0, 0,0,255, 0.0,0.0,0.0, 255,255,255);
}

int COrigami::defaultSetup( char *filename ){
	FILE *fp = fopen( filename, "w" );
	if( fp == NULL ) return 1;
	fprintf(fp, "VertexSize 4\n");
	fprintf(fp, " 10.0  10.0\n");
	fprintf(fp, "-10.0  10.0\n");
	fprintf(fp, "-10.0 -10.0\n");
	fprintf(fp, " 10.0 -10.0\n");
	fprintf(fp, "FrontColor  255   0   0\n");
	fprintf(fp, "BackColor     0 255 255\n");
	fprintf(fp, "BorderColor 255 255 255\n");
	fprintf(fp, "/FrontTexture  chiyo1.bmp\n");
	fprintf(fp, "/BackTexture   chiyo3.bmp\n");
	fclose( fp );
	return 0;
}

void COrigami::save( char *filename ){
	FILE *fp = fopen( filename, "wb" );
	if( fp  == NULL ){
		printf( "Can't create %s\n", filename );
		return;
	}
	for( int i=0; i<maxStage; i++ ) stage[i]->save( fp );
	fclose(fp);
}

void COrigami::load( char *filename ){
	FILE *fp = fopen( filename, "rb" );
	if( fp  == NULL ){
		printf( "Can't open %s\n", filename );
		return;
	}
	this->reset( 0 );
	while( stage[maxStage]->load( fp ) ){
		stage[maxStage]->renew();
		stage[maxStage+1] = new CStage( *stage[maxStage] );
		maxStage++;
	}
	fclose(fp);
}

void COrigami::reset( char *filename ){
	this->reset( 0 );
	if( ( animationFp = fopen( filename, "rb" ) ) == NULL ){
		printf( "Can't open %s\n", filename );
		return;
	}
	animationStage.load( animationFp );
	animationPosition = 0.0;
}

FoldResult COrigami::renew( double deltaPosition ){
	animationPosition += deltaPosition;
	FoldResult foldResult = stage[maxStage]->renew( animationStage, animationPosition );
	if( foldResult == FoldFailure ) return FoldFailure;
	stage[maxStage+1] = new CStage( *stage[maxStage] );
	maxStage++;
	if( foldResult == FoldTerminal ){
		animationPosition = 0.001;
		if( animationStage.load( animationFp ) == 0 ){
			fclose( animationFp );
			return FoldComplete;
		}
		return FoldTerminal;
	}
	return FoldProgress;
}
